/*=============================================================================
Blobby Volley 2
Copyright (C) 2006 Jonathan Sieber (jonathan_sieber@yahoo.de)
Copyright (C) 2006 Daniel Knobe (daniel-knobe@web.de)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
=============================================================================*/

#pragma once


#include <iosfwd>

#include "Global.h"
#include "GenericIOFwd.h"

struct GameLogicState
{
	unsigned int leftScore;
	unsigned int rightScore;
	unsigned int hitCount[MAX_PLAYERS];
	PlayerSide servingPlayer;
	PlayerSide winningPlayer = NO_PLAYER;

	unsigned int squish[MAX_PLAYERS];
	unsigned int squishWall;
	unsigned int squishGround;

	bool isGameRunning;
	bool isBallValid;


	void swapSides();

	void serialize(GenericOut* io) const;
	void deserialize(GenericIn* io);
};

std::ostream& operator<<(std::ostream&, const GameLogicState&);
